#include <cmath>
#include <msp/core/raii.h>
#include <msp/gl/extensions/arb_direct_state_access.h>
#include <msp/gl/extensions/arb_texture_storage.h>
#include <msp/gl/extensions/ext_texture3d.h>
#include <msp/gl/extensions/ext_texture_array.h>
#include <msp/graphics/image.h>
#include "bindable.h"
#include "error.h"
#include "pixelstore.h"
#include "texture3d.h"

using namespace std;

namespace Msp {
namespace GL {

Texture3D::Texture3D(GLenum t):
	Texture(t),
	width(0),
	height(0),
	depth(0),
	allocated(0)
{ }

Texture3D::Texture3D():
	Texture(GL_TEXTURE_3D),
	width(0),
	height(0),
	depth(0),
	allocated(0)
{
	static Require _req(EXT_texture3D);
}

void Texture3D::storage(PixelFormat fmt, unsigned wd, unsigned ht, unsigned dp, unsigned lv)
{
	if(width>0)
		throw invalid_operation("Texture3D::storage");
	if(wd==0 || ht==0 || dp==0)
		throw invalid_argument("Texture3D::storage");

	set_format(fmt);
	width = wd;
	height = ht;
	depth = dp;
	levels = get_n_levels();
	if(lv>0)
		levels = min(levels, lv);
}

void Texture3D::allocate(unsigned level)
{
	if(width==0 || height==0 || depth==0)
		throw invalid_operation("Texture3D::allocate");
	if(level>=levels)
		throw invalid_argument("Texture3D::allocate");
	if(allocated&(1<<level))
		return;

	if(ARB_texture_storage)
	{
		Conditional<BindRestore> _bind(!ARB_direct_state_access, this);
		if(ARB_direct_state_access)
			glTextureStorage3D(id, levels, storage_fmt, width, height, depth);
		else
			glTexStorage3D(target, levels, storage_fmt, width, height, depth);
		apply_swizzle();
		allocated |= (1<<levels)-1;
	}
	else
		image(level, 0);
}

void Texture3D::image(unsigned level, const void *data)
{
	if(width==0 || height==0 || depth==0)
		throw invalid_operation("Texture3D::image");

	unsigned w = width;
	unsigned h = height;
	unsigned d = depth;
	get_level_size(level, w, h, d);

	if(ARB_texture_storage)
		return sub_image(level, 0, 0, 0, w, h, d, data);

	BindRestore _bind(this);

	if(!allocated)
	{
		glTexParameteri(target, GL_TEXTURE_MAX_LEVEL, levels-1);
		apply_swizzle();
	}

	PixelComponents comp = get_components(storage_fmt);
	GLenum type = get_gl_type(get_component_type(storage_fmt));
	glTexImage3D(target, level, storage_fmt, width, height, depth, 0, comp, type, data);

	allocated |= 1<<level;
	if(auto_gen_mipmap && level==0)
	{
		generate_mipmap();
		allocated |= (1<<levels)-1;
	}
}

void Texture3D::image(unsigned level, PixelComponents comp, DataType type, const void *data)
{
	if(comp!=get_components(format) || type!=get_component_type(format))
		throw incompatible_data("Texture3D::image");
	image(level, data);
}

void Texture3D::sub_image(unsigned level, int x, int y, int z, unsigned wd, unsigned ht, unsigned dp, const void *data)
{
	if(width==0 || height==0 || depth==0)
		throw invalid_operation("Texture3D::image");

	Conditional<BindRestore> _bind(!ARB_direct_state_access, this);
	allocate(level);

	PixelComponents comp = get_components(storage_fmt);
	GLenum type = get_gl_type(get_component_type(storage_fmt));
	if(ARB_direct_state_access)
		glTextureSubImage3D(id, level, x, y, z, wd, ht, dp, comp, type, data);
	else
		glTexSubImage3D(target, level, x, y, z, wd, ht, dp, comp, type, data);

	if(auto_gen_mipmap && level==0)
		generate_mipmap();
}

void Texture3D::sub_image(unsigned level, int x, int y, int z, unsigned wd, unsigned ht, unsigned dp, PixelComponents comp, DataType type, const void *data)
{
	if(comp!=get_components(format) || type!=get_component_type(format))
		throw incompatible_data("Texture3D::sub_image");
	sub_image(level, x, y, z, wd, ht, dp, data);
}

void Texture3D::image(const Graphics::Image &img, unsigned lv)
{
	unsigned w = img.get_width();
	unsigned h = img.get_height();

	if(h%w)
		throw incompatible_data("Texture3D::load_image");
	unsigned d = h/w;
	h = w;

	PixelFormat fmt = pixelformat_from_image(img);
	if(width==0)
		storage(make_pixelformat(get_components(fmt), get_component_type(fmt), use_srgb_format), w, h, d, lv);
	else if(w!=width || h!=height || d!=depth)
		throw incompatible_data("Texture3D::load_image");

	PixelStore pstore = PixelStore::from_image(img);
	BindRestore _bind_ps(pstore);

	image(0, img.get_pixels());
}

unsigned Texture3D::get_n_levels() const
{
	unsigned s = max(width, height);
	if(target!=GL_TEXTURE_2D_ARRAY)
		s = max(s, depth);
	unsigned n = 0;
	for(; s; s>>=1, ++n) ;
	return n;
}

void Texture3D::get_level_size(unsigned level, unsigned &w, unsigned &h, unsigned &d) const
{
	w >>= level;
	h >>= level;
	if(target!=GL_TEXTURE_2D_ARRAY)
		d >>= level;

	if(!w && (h || d))
		w = 1;
	if(!h && (w || d))
		h = 1;
	if(!d && (w || h))
		d = 1;
}

UInt64 Texture3D::get_data_size() const
{
	return id ? width*height*depth*get_pixel_size(storage_fmt) : 0;
}


Texture3D::Loader::Loader(Texture3D &t):
	DataFile::DerivedObjectLoader<Texture3D, Texture::Loader>(t)
{
	init();
}

Texture3D::Loader::Loader(Texture3D &t, Collection &c):
	DataFile::DerivedObjectLoader<Texture3D, Texture::Loader>(t, c)
{
	init();
}

void Texture3D::Loader::init()
{
	add("raw_data", &Loader::raw_data);
	add("storage", &Loader::storage);
	add("storage", &Loader::storage_levels);
}

void Texture3D::Loader::raw_data(const string &data)
{
	obj.image(0, data.data());
}

void Texture3D::Loader::storage(PixelFormat fmt, unsigned w, unsigned h, unsigned d)
{
	obj.storage(fmt, w, h, d);
}

void Texture3D::Loader::storage_levels(PixelFormat fmt, unsigned w, unsigned h, unsigned d, unsigned l)
{
	obj.storage(fmt, w, h, d, l);
}

} // namespace GL
} // namespace Msp
