#include <msp/core/raii.h>
#include <msp/strings/format.h>
#include <msp/strings/regex.h>
#include <msp/strings/utils.h>
#include "builtin.h"
#include "glsl_error.h"
#include "parser.h"

#undef interface

using namespace std;

namespace Msp {
namespace GL {
namespace SL {

Parser::Parser():
	preprocessor(tokenizer),
	module(0)
{
	tokenizer.signal_preprocess.connect(sigc::mem_fun(&preprocessor, &Preprocessor::preprocess));
	preprocessor.signal_version.connect(sigc::mem_fun(this, &Parser::set_required_version));
	preprocessor.signal_source.connect(sigc::mem_fun(this, &Parser::source_reference));
	preprocessor.signal_stage_change.connect(sigc::mem_fun(this, &Parser::stage_change));
	preprocessor.signal_line.connect(sigc::mem_fun(this, &Parser::line_change));
}

Parser::~Parser()
{
	delete module;
}

Module &Parser::parse(const string &s, const string &n, int i)
{
	source = s;
	parse_source(n, i);
	return *module;
}

Module &Parser::parse(IO::Base &io, const string &n, int i)
{
	source = string();
	while(!io.eof())
	{
		char buffer[4096];
		unsigned len = io.read(buffer, sizeof(buffer));
		source.append(buffer, len);
	}
	parse_source(n, i);
	return *module;
}

void Parser::parse_source(const string &name, int index)
{
	delete module;
	module = new Module;

	cur_stage = &module->shared;
	base_index = index;
	source_index = index;
	if(index>=0)
		source_reference(1, name);

	// TODO Need to somehow get type names from imports
	if(const Stage *builtin = get_builtins(Stage::SHARED))
	{
		for(map<string, TypeDeclaration *>::const_iterator i=builtin->types.begin(); i!=builtin->types.end(); ++i)
			declared_types.insert(i->first);
	}
	else
	{
		declared_types.insert("void");
		declared_types.insert("bool");
		declared_types.insert("int");
		declared_types.insert("float");
	}

	tokenizer.begin(source, name);
	allow_stage_change = true;
	while(!tokenizer.peek_token().empty())
		if(RefPtr<Statement> statement = parse_with_recovery(&Parser::parse_global_declaration))
			cur_stage->content.body.push_back(statement);

	if(!errors.empty())
		throw invalid_shader_source(join(errors.begin(), errors.end(), "\n"));
}

void Parser::set_required_version(const Version &ver)
{
	cur_stage->required_features.glsl_version = ver;
}

void Parser::source_reference(unsigned index, const string &name)
{
	if(index<1 || base_index<0)
		throw invalid_shader_source(tokenizer.get_location(), "Invalid source reference");

	module->source_map.set_name(base_index+index-1, name);
}

void Parser::stage_change(Stage::Type stage)
{
	if(!allow_stage_change)
		throw invalid_shader_source(tokenizer.get_location(), "Changing stage not allowed here");
	else if(stage<=cur_stage->type)
		throw invalid_shader_source(tokenizer.get_location(), "Stage '%s' not allowed here", Stage::get_stage_name(stage));

	module->stages.push_back(stage);

	if(cur_stage->type!=Stage::SHARED)
		module->stages.back().previous = cur_stage;
	cur_stage = &module->stages.back();
}

void Parser::line_change(int index, unsigned line)
{
	if(index>0)
		source_index = base_index+index-1;
	else if(index==0)
		source_index = 0;
	else
		index = source_index;

	string name = module->source_map.get_name(index);
	if(name.empty())
		name = format("<%d>", index);
	tokenizer.set_location(Location(name, line));
}

string Parser::expect_type()
{
	string token = tokenizer.parse_token();
	if(!is_type(token))
		throw parse_error(tokenizer.get_location(), token, "a type");
	return token;
}

string Parser::expect_identifier()
{
	string token = tokenizer.parse_token();
	if(!is_identifier(token))
		throw parse_error(tokenizer.get_location(), token, "an identifier");
	return token;
}

int Parser::expect_integer()
{
	string token = tokenizer.parse_token();
	if(!isnumrc(token))
		throw parse_error(tokenizer.get_location(), token, "an integer literal");
	return lexical_cast<int>(token);
}

bool Parser::check(const string &token)
{
	bool result = (tokenizer.peek_token()==token);
	if(result)
		tokenizer.parse_token();
	return result;
}

bool Parser::is_interface_qualifier(const string &token)
{
	return (token=="uniform" || token=="in" || token=="out");
}

bool Parser::is_sampling_qualifier(const string &token)
{
	return (token=="centroid" || token=="sample");
}

bool Parser::is_interpolation_qualifier(const string &token)
{
	return (token=="smooth" || token=="flat" || token=="noperspective");
}

bool Parser::is_precision_qualifier(const string &token)
{
	return (token=="highp" || token=="mediump" || token=="lowp");
}

bool Parser::is_qualifier(const string &token)
{
	return (token=="const" ||
		is_interface_qualifier(token) ||
		is_sampling_qualifier(token) ||
		is_interpolation_qualifier(token) ||
		is_precision_qualifier(token));
}

bool Parser::is_type(const string &token)
{
	return declared_types.count(token);
}

bool Parser::is_identifier(const string &token)
{
	static Regex re("^[a-zA-Z_][a-zA-Z0-9_]*$");
	return re.match(token);
}

template<typename T>
RefPtr<T> Parser::create_node()
{
	RefPtr<T> node = new T;
	node->source = source_index;
	node->line = tokenizer.get_location().line;
	return node;
}

template<typename T>
RefPtr<T> Parser::parse_with_recovery(RefPtr<T> (Parser::*parse_func)())
{
	tokenizer.clear_progress_mark();
	try
	{
		return (this->*parse_func)();
	}
	catch(const invalid_shader_source &exc)
	{
		errors.push_back(exc.what());
	}

	if(tokenizer.get_last_token()!=";" || !tokenizer.get_progress_mark())
	{
		unsigned scope_level = 0;
		while(1)
		{
			if(tokenizer.peek_token()=="}" && scope_level==0)
			{
				if(!tokenizer.get_progress_mark())
					tokenizer.parse_token();
				break;
			}

			string token = tokenizer.parse_token();
			if(token=="}")
			{
				--scope_level;
				if(scope_level==0)
					break;
			}
			else if(token=="{")
				++scope_level;
			else if(token==";" && scope_level==0)
				break;
			else if(token.empty())
				break;
		}
	}

	return RefPtr<T>();
}

RefPtr<Statement> Parser::parse_global_declaration()
{
	string token = tokenizer.peek_token();
	SetFlag disallow(allow_stage_change, false);

	if(token=="import")
		return parse_import();
	else if(token=="precision")
		return parse_precision();
	else if(token=="layout")
	{
		RefPtr<Layout> layout = parse_layout();
		token = tokenizer.peek_token();
		if(is_interface_qualifier(token) && tokenizer.peek_token(1)==";")
		{
			RefPtr<InterfaceLayout> iface_lo = create_node<InterfaceLayout>();
			iface_lo->layout.source = layout->source;
			iface_lo->layout.line = layout->line;
			iface_lo->layout.qualifiers = layout->qualifiers;
			iface_lo->interface = tokenizer.parse_token();
			tokenizer.expect(";");
			return iface_lo;
		}
		else if(is_interface_qualifier(token) && tokenizer.peek_token(2)=="{")
		{
			RefPtr<InterfaceBlock> iface = parse_interface_block();
			iface->layout = layout;
			return iface;
		}
		else
		{
			RefPtr<VariableDeclaration> var = parse_variable_declaration();
			var->layout = layout;
			return var;
		}
	}
	else if(token=="typedef")
		return parse_type_declaration();
	else if(token=="struct")
		return parse_struct_declaration();
	else if(is_interface_qualifier(token))
	{
		string next = tokenizer.peek_token(1);
		if(is_type(next) || is_qualifier(next))
			return parse_variable_declaration();
		else
			return parse_interface_block();
	}
	else if(is_qualifier(token))
		return parse_variable_declaration();
	else if(token=="virtual")
		return parse_function_declaration();
	else if(is_type(token))
	{
		if(tokenizer.peek_token(2)=="(")
			return parse_function_declaration();
		else
			return parse_variable_declaration();
	}
	else if(token.empty())
		return 0;
	else
		throw parse_error(tokenizer.get_location(), token, "a global declaration");
}

RefPtr<Statement> Parser::parse_statement()
{
	string token = tokenizer.peek_token();
	if(token=="if")
		return parse_conditional();
	else if(token=="for")
		return parse_for();
	else if(token=="while")
		return parse_while();
	else if(token=="passthrough")
		return parse_passthrough();
	else if(token=="return")
		return parse_return();
	else if(token=="break" || token=="continue" || token=="discard")
	{
		RefPtr<Jump> jump = create_node<Jump>();
		jump->keyword = tokenizer.parse_token();
		tokenizer.expect(";");

		return jump;
	}
	else if(is_qualifier(token) || is_type(token))
		return parse_variable_declaration();
	else if(token==";")
	{
		tokenizer.parse_token();
		throw invalid_shader_source(tokenizer.get_location(), "Empty statement not allowed");
	}
	else if(!token.empty())
	{
		RefPtr<ExpressionStatement> expr = create_node<ExpressionStatement>();
		expr->expression = parse_expression();
		tokenizer.expect(";");

		return expr;
	}
	else
		throw parse_error(tokenizer.get_location(), token, "a statement");
}

RefPtr<Import> Parser::parse_import()
{
	if(cur_stage->type!=Stage::SHARED)
		throw invalid_shader_source(tokenizer.get_location(), "Imports are only allowed in the shared section");

	tokenizer.expect("import");
	RefPtr<Import> import = create_node<Import>();
	import->module = expect_identifier();
	tokenizer.expect(";");
	return import;
}

RefPtr<Precision> Parser::parse_precision()
{
	tokenizer.expect("precision");
	RefPtr<Precision> precision = create_node<Precision>();

	precision->precision = tokenizer.parse_token();
	if(!is_precision_qualifier(precision->precision))
		throw parse_error(tokenizer.get_location(), precision->precision, "a precision qualifier");

	// TODO Add validation for this
	precision->type = expect_type();

	tokenizer.expect(";");

	return precision;
}

RefPtr<Layout> Parser::parse_layout()
{
	tokenizer.expect("layout");
	tokenizer.expect("(");
	RefPtr<Layout> layout = create_node<Layout>();
	while(1)
	{
		string token = tokenizer.parse_token();
		if(token==")")
			throw parse_error(tokenizer.get_location(), token, "a layout qualifier name");

		layout->qualifiers.push_back(token);
		Layout::Qualifier &qual = layout->qualifiers.back();

		if((qual.has_value = check("=")))
		{
			if(qual.name=="constant_id" && tokenizer.peek_token()=="auto")
			{
				qual.value = -1;
				tokenizer.parse_token();
			}
			else
				qual.value = expect_integer();
		}

		if(tokenizer.peek_token()==")")
			break;

		tokenizer.expect(",");
	}
	tokenizer.expect(")");

	return layout;
}

template<typename T>
void Parser::parse_block(Block &block, bool require_braces, RefPtr<T> (Parser::*parse_content)())
{
	bool have_braces = (require_braces || tokenizer.peek_token()=="{");
	if(have_braces)
		tokenizer.expect("{");

	if(have_braces)
	{
		while(tokenizer.peek_token()!="}")
			if(RefPtr<Statement> node = parse_with_recovery(parse_content))
				block.body.push_back(node);
	}
	else
		block.body.push_back((this->*parse_content)());

	block.use_braces = (require_braces || block.body.size()!=1);

	if(have_braces)
		tokenizer.expect("}");
}

RefPtr<Expression> Parser::parse_expression(const Operator *outer_oper)
{
	unsigned outer_precedence = (outer_oper ? outer_oper->precedence+(outer_oper->assoc==Operator::RIGHT_TO_LEFT) : 20);
	RefPtr<Expression> left;
	VariableReference *left_var = 0;
	while(1)
	{
		string token = tokenizer.peek_token();

		const Operator *oper = 0;
		for(const Operator *i=Operator::operators; (!oper && i->type); ++i)
			if(token==i->token && (!left || i->type!=Operator::PREFIX) && (left || i->type!=Operator::POSTFIX))
				oper = i;

		bool lower_precedence = (oper && oper->type!=Operator::PREFIX && oper->precedence>=outer_precedence);
		if(token==";" || token==")" || token=="]" || token=="," || token==":" || lower_precedence)
		{
			if(left)
				return left;
			else
				throw parse_error(tokenizer.get_location(), token, "an expression");
		}
		else if(left)
		{
			if(token=="(")
			{
				if(!left_var)
					throw invalid_shader_source(tokenizer.get_location(), "Syntax error before '(': function name must be an identifier");
				left = parse_function_call(*left_var);
			}
			else if(token==".")
			{
				RefPtr<MemberAccess> memacc = create_node<MemberAccess>();
				memacc->left = left;
				memacc->oper = oper;
				tokenizer.parse_token();
				memacc->member = expect_identifier();
				left = memacc;
			}
			else if(oper && oper->type==Operator::POSTFIX)
			{
				RefPtr<UnaryExpression> unary = create_node<UnaryExpression>();
				unary->oper = oper;
				tokenizer.parse_token();
				unary->expression = left;
				left = unary;
			}
			else if(oper && oper->type==Operator::BINARY)
				left = parse_binary(left, *oper);
			else if(oper && oper->type==Operator::TERNARY)
				left = parse_ternary(left, *oper);
			else
				throw parse_error(tokenizer.get_location(), token, "an operator");
			left_var = 0;
		}
		else
		{
			if(token=="(")
			{
				tokenizer.parse_token();
				left = parse_expression();
				tokenizer.expect(")");
			}
			else if(isdigit(token[0]) || token=="true" || token=="false")
				left = parse_literal();
			else if(is_identifier(token))
			{
				RefPtr<VariableReference> var = create_node<VariableReference>();
				var->name = expect_identifier();
				left = var;
				left_var = var.get();
			}
			else if(oper && oper->type==Operator::PREFIX)
			{
				RefPtr<UnaryExpression> unary = create_node<UnaryExpression>();
				unary->oper = oper;
				tokenizer.parse_token();
				unary->expression = parse_expression(oper);
				left = unary;
			}
			else
				throw parse_error(tokenizer.get_location(), token, "an expression");
		}
	}
}

RefPtr<Literal> Parser::parse_literal()
{
	RefPtr<Literal> literal = create_node<Literal>();
	literal->token = tokenizer.parse_token();
	if(isdigit(literal->token[0]))
	{
		// TODO have the tokenizer return the type of the token
		if(isnumrc(literal->token))
			literal->value = lexical_cast<int>(literal->token);
		else
			literal->value = lexical_cast<float>(literal->token);
	}
	else if(literal->token=="true" || literal->token=="false")
		literal->value = (literal->token=="true");
	else
		throw parse_error(tokenizer.get_location(), literal->token, "a literal");

	return literal;
}

RefPtr<BinaryExpression> Parser::parse_binary(const RefPtr<Expression> &left, const Operator &oper)
{
	RefPtr<BinaryExpression> binary = (oper.precedence==16 ?
		static_cast<RefPtr<BinaryExpression> >(create_node<Assignment>()) : create_node<BinaryExpression>());
	binary->left = left;
	binary->oper = &oper;
	tokenizer.expect(oper.token);
	if(oper.token2[0])
	{
		binary->right = parse_expression();
		tokenizer.expect(oper.token2);
	}
	else
		binary->right = parse_expression(&oper);
	return binary;
}

RefPtr<TernaryExpression> Parser::parse_ternary(const RefPtr<Expression> &cond, const Operator &oper)
{
	RefPtr<TernaryExpression> ternary = create_node<TernaryExpression>();
	ternary->condition = cond;
	ternary->oper = &oper;
	tokenizer.expect("?");
	ternary->true_expr = parse_expression(&oper);
	tokenizer.expect(":");
	ternary->false_expr = parse_expression(&oper);
	return ternary;
}

RefPtr<FunctionCall> Parser::parse_function_call(const VariableReference &var)
{
	RefPtr<FunctionCall> call = create_node<FunctionCall>();
	call->name = var.name;
	call->oper = &Operator::get_operator("(", Operator::POSTFIX);
	tokenizer.expect("(");
	while(tokenizer.peek_token()!=")")
	{
		if(!call->arguments.empty())
			tokenizer.expect(",");
		call->arguments.push_back(parse_expression());
	}
	tokenizer.expect(")");
	return call;
}

RefPtr<TypeDeclaration> Parser::parse_type_declaration()
{
	tokenizer.expect("typedef");

	RefPtr<TypeDeclaration> type;
	if(tokenizer.peek_token()=="image")
		type = parse_image_type_declaration();
	else
		type = parse_basic_type_declaration();

	tokenizer.expect(";");
	declared_types.insert(type->name);
	return type;
}

RefPtr<BasicTypeDeclaration> Parser::parse_basic_type_declaration()
{
	RefPtr<BasicTypeDeclaration> type = create_node<BasicTypeDeclaration>();

	if(tokenizer.peek_token()=="vector")
	{
		type->kind = BasicTypeDeclaration::VECTOR;

		tokenizer.parse_token();
		tokenizer.expect("(");
		type->size = expect_integer();
		tokenizer.expect(")");
	}

	type->base = expect_type();
	type->name = expect_identifier();

	if(type->kind==BasicTypeDeclaration::ALIAS && check("["))
	{
		type->kind = BasicTypeDeclaration::ARRAY;
		tokenizer.expect("]");
	}

	return type;
}

RefPtr<ImageTypeDeclaration> Parser::parse_image_type_declaration()
{
	tokenizer.expect("image");
	tokenizer.expect("(");

	RefPtr<ImageTypeDeclaration> type = create_node<ImageTypeDeclaration>();
	while(1)
	{
		string token = tokenizer.parse_token();
		if(token=="dimensions")
		{
			tokenizer.expect("=");
			token = tokenizer.parse_token();
			if(token=="1")
				type->dimensions = ImageTypeDeclaration::ONE;
			else if(token=="2")
				type->dimensions = ImageTypeDeclaration::TWO;
			else if(token=="3")
				type->dimensions = ImageTypeDeclaration::THREE;
			else if(token=="cube")
				type->dimensions = ImageTypeDeclaration::CUBE;
			else
				throw parse_error(tokenizer.get_location(), token, "dimensions");

			if(check("["))
			{
				type->array = true;
				tokenizer.expect("]");
			}
		}
		else if(token=="sampled")
			type->sampled = true;
		else if(token=="shadow")
			type->shadow = true;
		else
			throw parse_error(tokenizer.get_location(), token, "image type attribute");

		token = tokenizer.peek_token();
		if(token==")")
			break;

		tokenizer.expect(",");
	}
	tokenizer.expect(")");

	type->base = expect_type();
	type->name = expect_identifier();

	return type;
}

RefPtr<StructDeclaration> Parser::parse_struct_declaration()
{
	tokenizer.expect("struct");
	RefPtr<StructDeclaration> strct = create_node<StructDeclaration>();

	strct->name = expect_identifier();
	parse_block(strct->members, true, &Parser::parse_variable_declaration);
	tokenizer.expect(";");

	declared_types.insert(strct->name);
	return strct;
}

RefPtr<VariableDeclaration> Parser::parse_variable_declaration()
{
	RefPtr<VariableDeclaration> var = create_node<VariableDeclaration>();

	string token = tokenizer.peek_token();
	while(is_qualifier(token))
	{
		tokenizer.parse_token();
		if(is_interface_qualifier(token))
			var->interface = token;
		else if(is_sampling_qualifier(token))
			var->sampling = token;
		else if(is_interpolation_qualifier(token))
			var->interpolation = token;
		else if(is_precision_qualifier(token))
			var->precision = token;
		else if(token=="const")
			var->constant = true;
		token = tokenizer.peek_token();
	}

	var->type = expect_type();
	var->name = expect_identifier();

	if(check("["))
	{
		var->array = true;
		if(!check("]"))
		{
			var->array_size = parse_expression();
			tokenizer.expect("]");
		}
	}

	if(check("="))
		var->init_expression = parse_expression();

	tokenizer.expect(";");
	return var;
}

RefPtr<VariableDeclaration> Parser::parse_variable_declaration_with_layout()
{
	RefPtr<Layout> layout;
	if(tokenizer.peek_token()=="layout")
		layout = parse_layout();

	RefPtr<VariableDeclaration> var = parse_variable_declaration();
	var->layout = layout;

	return var;
}

RefPtr<FunctionDeclaration> Parser::parse_function_declaration()
{
	RefPtr<FunctionDeclaration> func = create_node<FunctionDeclaration>();

	func->virtua = check("virtual");
	func->return_type = expect_type();
	func->name = expect_identifier();
	tokenizer.expect("(");
	while(tokenizer.peek_token()!=")")
	{
		if(!func->parameters.empty())
			tokenizer.expect(",");

		RefPtr<VariableDeclaration> var = create_node<VariableDeclaration>();
		string token = tokenizer.peek_token();
		if(token=="in" || token=="out" || token=="inout")
			var->interface = tokenizer.parse_token();
		var->type = expect_type();
		var->name = expect_identifier();
		func->parameters.push_back(var);
	}
	tokenizer.expect(")");

	func->overrd = check("override");

	string token = tokenizer.peek_token();
	if(token=="{")
	{
		func->definition = func.get();
		parse_block(func->body, true, &Parser::parse_statement);
	}
	else if(token==";")
		tokenizer.parse_token();
	else
		throw parse_error(tokenizer.get_location(), token, "'{' or ';'");

	return func;
}

RefPtr<InterfaceBlock> Parser::parse_interface_block()
{
	RefPtr<InterfaceBlock> iface = create_node<InterfaceBlock>();

	iface->interface = tokenizer.parse_token();
	if(!is_interface_qualifier(iface->interface))
		throw parse_error(tokenizer.get_location(), iface->interface, "an interface qualifier");

	iface->block_name = expect_identifier();
	iface->members = new Block;
	parse_block(*iface->members, true, &Parser::parse_variable_declaration_with_layout);
	if(!check(";"))
	{
		iface->instance_name = expect_identifier();
		if(check("["))
		{
			iface->array = true;
			tokenizer.expect("]");
		}
		tokenizer.expect(";");
	}

	return iface;
}

RefPtr<Conditional> Parser::parse_conditional()
{
	tokenizer.expect("if");
	RefPtr<Conditional> cond = create_node<Conditional>();
	tokenizer.expect("(");
	cond->condition = parse_expression();
	tokenizer.expect(")");

	parse_block(cond->body, false, &Parser::parse_statement);

	string token = tokenizer.peek_token();
	if(token=="else")
	{
		tokenizer.parse_token();
		parse_block(cond->else_body, false, &Parser::parse_statement);
	}

	return cond;
}

RefPtr<Iteration> Parser::parse_for()
{
	tokenizer.expect("for");
	RefPtr<Iteration> loop = create_node<Iteration>();
	tokenizer.expect("(");
	string token = tokenizer.peek_token();
	if(is_type(token))
		loop->init_statement = parse_statement();
	else
	{
		if(token!=";")
		{
			RefPtr<ExpressionStatement> expr = create_node<ExpressionStatement>();
			expr->expression = parse_expression();
			loop->init_statement = expr;
		}
		tokenizer.expect(";");
	}
	if(tokenizer.peek_token()!=";")
		loop->condition = parse_expression();
	tokenizer.expect(";");
	if(tokenizer.peek_token()!=")")
		loop->loop_expression = parse_expression();
	tokenizer.expect(")");

	parse_block(loop->body, false, &Parser::parse_statement);

	return loop;
}

RefPtr<Iteration> Parser::parse_while()
{
	tokenizer.expect("while");
	RefPtr<Iteration> loop = create_node<Iteration>();
	tokenizer.expect("(");
	loop->condition = parse_expression();
	tokenizer.expect(")");

	parse_block(loop->body, false, &Parser::parse_statement);

	return loop;
}

RefPtr<Passthrough> Parser::parse_passthrough()
{
	tokenizer.expect("passthrough");
	RefPtr<Passthrough> pass = create_node<Passthrough>();
	if(cur_stage->type==Stage::GEOMETRY)
	{
		tokenizer.expect("[");
		pass->subscript = parse_expression();
		tokenizer.expect("]");
	}
	tokenizer.expect(";");
	return pass;
}

RefPtr<Return> Parser::parse_return()
{
	tokenizer.expect("return");
	RefPtr<Return> ret = create_node<Return>();
	if(tokenizer.peek_token()!=";")
		ret->expression = parse_expression();
	tokenizer.expect(";");
	return ret;
}

} // namespace SL
} // namespace GL
} // namespace Msp
